//
// jja: swiss army knife for chess file formats
// src/merge.rs: PolyGlot Book Merging
//
// Copyright (c) 2023 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use crate::polyglot::BookEntry;

/// Represents various strategies for merging two Polyglot opening books.
///
/// When merging two opening books, different strategies can be used to determine
/// how to combine the weights of the same moves in different books.
#[derive(Copy, Clone, Debug, PartialEq)]
pub enum MergeStrategy {
    /// Calculates the average weight of the moves in both books.
    AvgWeight,
    /// Takes the maximum weight of the moves in both books.
    MaxWeight,
    /// Takes the minimum weight of the moves in both books.
    MinWeight,
    /// Prioritizes the moves of the first book, ignoring moves from the second book.
    Ours,
    /// Merges by computing the weighted average of percentage weights, taking into account the
    /// total number of entries in each book. This approach gives higher importance to moves from
    /// larger books versus smaller ones, ensuring that the resultant weights reflect the relative
    /// contributions of each book based on its size.
    PercentageAverage,
    /// Calculates the sum of the weights of the moves in both books.
    SumWeight,
    /// Calculates the weighted average weight of the moves in both books, considering the given
    /// weights for each book.
    ///
    /// The tuple contains the weights associated with the first and second books, respectively.
    WeightedAverageWeight(f64, f64),
    /// Geometric scaling strategy: The geometric scale focuses on multiplying numbers together
    /// rather than adding, which can help in equalizing disparities.
    GeometricScaling,
    /// Calculates the harmonic mean weight of the moves in both books. This approach tends to
    /// favor more balanced weights and is less influenced by extreme values.
    HarmonicMean,
    /// Remap the weights in a non-linear fashion using the sigmoid function. The idea here is to
    /// diminish the influence of extreme values, which might be causing the dissatisfaction in
    /// previous strategies.
    Sigmoid,
    /// Calculates weight based on relative position in sorted move entries.
    Sort,
    // Add more strategies here
}

/// Calculates the average of the weights of two book entries.
///
/// # Arguments
///
/// * `entry1` - The first book entry.
/// * `entry2` - The second book entry.
///
/// # Returns
///
/// * A `u16` representing the average weight of the two book entries.
///   In case of overflow, it returns `u16::MAX`.
pub fn average_weight(entry1: &BookEntry, entry2: &BookEntry) -> u16 {
    let entry1_weight: u32 = entry1.weight.into();
    let entry2_weight: u32 = entry2.weight.into();
    let sum = entry1_weight.saturating_add(entry2_weight);
    (sum / 2).try_into().unwrap_or(u16::MAX)
}

/// Calculates the weighted average of the weights of two book entries,
/// considering the given weights for each book.
///
/// # Arguments
///
/// * `entry1` - The first book entry.
/// * `entry2` - The second book entry.
/// * `book1_weight` - The weight associated with the first book.
/// * `book2_weight` - The weight associated with the second book.
///
/// # Returns
///
/// * A `u16` representing the weighted average weight of the two book entries,
///   taking into account the weights of each book.
///   In case of overflow, it returns `u16::MAX`.
pub fn weighted_average_weight(
    entry1: &BookEntry,
    entry2: &BookEntry,
    book1_weight: f64,
    book2_weight: f64,
) -> u16 {
    let total_weight = book1_weight + book2_weight;
    let normalized_book1_weight = book1_weight / total_weight;
    let normalized_book2_weight = book2_weight / total_weight;

    let entry1_weight: f64 = entry1.weight.into();
    let entry2_weight: f64 = entry2.weight.into();

    let weighted_sum =
        entry1_weight * normalized_book1_weight + entry2_weight * normalized_book2_weight;
    weighted_sum.min(f64::from(u16::MAX)).round() as u16
}

/// Calculates the sigmoid of a number.
///
/// The sigmoid function maps any number to a value between 0 and 1.
///
/// # Arguments
///
/// * `x` - The input number.
///
/// # Returns
///
/// A f64 value between 0 and 1.
fn sigmoid(x: f64) -> f64 {
    1.0 / (1.0 + (-x).exp())
}

/// Calculates the inverse of the sigmoid function.
///
/// # Arguments
///
/// * `y` - The input number between 0 and 1.
///
/// # Returns
///
/// A f64 value that is the inverse of the sigmoid function.
fn inverse_sigmoid(y: f64) -> f64 {
    (y / (1.0 - y)).ln()
}

/// Merges two u16 weights using a non-linear sigmoid-based strategy.
///
/// # Arguments
///
/// * `a` - The first weight.
/// * `b` - The second weight.
///
/// # Returns
///
/// A u16 value that is the merged weight of the two input weights.
pub fn sigmoid_merge(a: u16, b: u16) -> u16 {
    // Normalize weights
    let a_norm = sigmoid(f64::from(a) - 32768.0);
    let b_norm = sigmoid(f64::from(b) - 32768.0);

    // Compute average
    let m_norm = (a_norm + b_norm) / 2.0;

    // Denormalize
    let merged = 32768.0 + inverse_sigmoid(m_norm);

    merged.round() as u16
}

/// Merges two u16 weights using the harmonic mean strategy.
///
/// # Arguments
///
/// * `a` - The first weight.
/// * `b` - The second weight.
///
/// # Returns
///
/// A u16 value that is the merged weight of the two input weights.
pub fn harmonic_merge(a: u16, b: u16) -> u16 {
    // If either weight is zero, handle accordingly.
    // In this implementation, we'll simply return the non-zero weight.
    if a == 0 {
        return b;
    }
    if b == 0 {
        return a;
    }

    let merged = 2.0 / (1.0 / f64::from(a) + 1.0 / f64::from(b));

    merged.round() as u16
}

/// Merges two u16 weights using the geometric scaling strategy.
///
/// # Arguments
///
/// * `a` - The first weight.
/// * `b` - The second weight.
///
/// # Returns
///
/// A u16 value that is the merged weight of the two input weights.
pub fn geometric_scaling_merge(a: u16, b: u16) -> u16 {
    // Scale the weights
    let a_scaled = (f64::from(a) / f64::from(u16::MAX)) * f64::from(u32::MAX);
    let b_scaled = (f64::from(b) / f64::from(u16::MAX)) * f64::from(u32::MAX);

    // Compute geometric mean in the scaled space
    let m_scaled = (a_scaled * b_scaled).sqrt();

    // Rescale to u16 range
    let merged = (m_scaled / f64::from(u32::MAX)) * f64::from(u16::MAX);

    merged.round() as u16
}
